"""
This package contains user utilities, structural subroutines and backend functions.

User Utilities
^^^^^^^^^^^^^^
- :mod:`Geometry <blueprints.utils.geometry>`
- :mod:`Perlin Noise <blueprints.utils.perlin>`

Structural Subroutines
^^^^^^^^^^^^^^^^^^^^^^
- :mod:`Views <blueprints.utils.view>`
- :mod:`Lattice <blueprints.utils.lattice>`

Backend Functions
^^^^^^^^^^^^^^^^^
- :mod:`Type Safty <blueprints.utils.typechecker>`
- :mod:`Naming <blueprints.utils.naming>`
- :mod:`Register <blueprints.utils.register>`
"""